/*
 * @(#) JotmRecovery.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 *
 * This module was originally developed by 
 *
 *  - Bull S.A. as part of the JOnAS application server code released in 
 *    July 1999 (www.bull.com)
 * 
 * --------------------------------------------------------------------------
 *  The original code and portions created by Bull SA are 
 *  Copyright (c) 1999 BULL SA  
 *  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * --------------------------------------------------------------------------
 * $Id: JotmRecovery.java,v 1.9 2005-09-14 22:23:57 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jotm;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Vector;
import java.util.List;
import java.util.Iterator;

import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.objectweb.howl.log.xa.XALogRecord;
import org.objectweb.howl.log.xa.XACommittingTx;
import javax.transaction.xa.XAException;

/**
 * 
 * @author Tony Ortiz
 */

/**
 * Thread class used to manage Resource Manager Registration.
 */
class processResourceManager extends Thread {
    private ThreadGroup xaResourceThreadGroup = null;
    Vector txrecovered = null;
    Vector rminfo = null;
    Vector rmreg = null;
    int rmsize = 0;
    int rmindex = 0;
    
    processResourceManager (ThreadGroup prmThreadGroup, Vector pvtxrecovered, Vector prminfo, Vector prmreg, int prmsize, int prmindex) {
        xaResourceThreadGroup = new ThreadGroup(prmThreadGroup, "XAResourceTG");
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("new processResourceManager");
        }
       
        txrecovered = pvtxrecovered;
        rminfo = prminfo;
        rmreg = prmreg;
        rmsize = prmsize;
        rmindex = prmindex;;
    }
    
    public void run() {
       if (TraceTm.recovery.isDebugEnabled()) {
           TraceTm.recovery.debug("thread for processResourceManager");
       }
    
       XAResource myregxares = null;
       RecoverRmInfo myrecoverRmInfo = null;
       String myrm = null;
       String myregrm = null;
       byte[] byxares = null;
       // int rcflag = 0;
    
       RmRegistration myrmreg = (RmRegistration) rmreg.elementAt(rmindex);
       myregrm = myrmreg.rmGetName();

       for (int j=0; j<rmsize; j++) {
           myrecoverRmInfo = (RecoverRmInfo) rminfo.elementAt(j);
           myrm = myrecoverRmInfo.getRecoverRm();
           byxares = myrecoverRmInfo.getRecoverXaRes();
            
           if (myregrm.equals(myrm)){
               // checkout the XAResource, this will "lock" the XAResource
               // we can only "unlock" the XAResource after all the threads
               // used to resolve (commit/rollback) the XAResource's XIDs
               // have successfully completed

               try {
                   myregxares = myrmreg.rmCheckoutXARes();
               } catch (XAException e) {
                   TraceTm.recovery.error("rmCheckoutXARes call failed during recovery " + e.getMessage());
                   myrmreg.rmCheckinXARes();
                   break;   
               }
            
               if (TraceTm.recovery.isDebugEnabled()) {
                   TraceTm.recovery.debug("Registered Resource Manager " + myregrm);
                   TraceTm.recovery.debug("Registered XAResource " + myregxares);
                   TraceTm.recovery.debug("Recover Resource Manager " + myrm);
                   TraceTm.recovery.debug("Recover XAResource " + new String(byxares));
               }
            
               // Xid [] javaxid = null;
               List recoveredXidList = new java.util.LinkedList();
                
               try {
                   // recover returns javax.transaction.xa.Xid objects
               //     javaxid = myregxares.recover(rcflag);
                   boolean first = true;
                   while (true) {
                       javax.transaction.xa.Xid[] javaxids;
                       javaxids = myregxares.recover(first ? XAResource.TMSTARTRSCAN : XAResource.TMNOFLAGS);
                       if (javaxids == null || javaxids.length == 0)
                           break;
                 
                       first = false;
                       recoveredXidList.addAll(Arrays.asList(javaxids));
                   }
               } catch (XAException e) {
                   TraceTm.recovery.error("xaResource.recover call failed during recovery " + e.getMessage());
                   myrmreg.rmCheckinXARes();
                   break;
               }
                
               // if (javaxid == null) {
               //     if (TraceTm.recovery.isDebugEnabled()) {
               //         TraceTm.recovery.debug("No XIDs to recover for Xares javaxid is null");
               //     }
               //     myrmreg.rmCheckinXARes();
               //     break; 
               // }
                
               if (TraceTm.recovery.isDebugEnabled()) {
                   // TraceTm.recovery.debug("javaxid size= " + javaxid.length);
                TraceTm.recovery.debug("recoveredXidList size= " + recoveredXidList.size());
               }
                
               // if (javaxid.length == 0) {
               if (recoveredXidList.size() == 0) {
                   if (TraceTm.recovery.isDebugEnabled()) {
                       TraceTm.recovery.debug("No XIDs to recover for Xares= "+ myregxares);
                   }
                   myrmreg.rmCheckinXARes();
                   break;
               }
               
               // set the action required for the XIDs located in our
               // txxidRecovered (array), pointed to by TxRecovered (vector)
                    
               // settxxidrecoveraction (myregxares, javaxid);
               settxxidrecoveraction(myregxares, recoveredXidList);
                    
               // We can now walk through TxRecovered and perform the action on
               // each of the Xids stored in TxxidRecovered.
            
               doActionXidRecover ();

               for (int w=0; w<10; w++) {  // only do it 10 times)
                   boolean xaresinthread = false;
                   int numthreads = xaResourceThreadGroup.activeCount();
                   
                   if (TraceTm.recovery.isDebugEnabled()) {
                       TraceTm.recovery.debug("xaResource active count= "+ numthreads);
                   }
                   
                   if (numthreads == 0) {
                       break;
                   }
                    
                   Thread[] listOfThreads = new Thread[numthreads];
                   xaResourceThreadGroup.enumerate(listOfThreads);

                   for (int i = 0; i < numthreads; i++) {
                       if (listOfThreads[i] != null) {
                           if (TraceTm.recovery.isDebugEnabled()) {
                               TraceTm.recovery.debug("xaResource in thread= "+ listOfThreads[i].getName());
                           }
                           
                           try {
                               Thread.sleep(100);  // 100 millisecond
                           } catch (InterruptedException e) {
                               ;
                           }
                           xaresinthread = true;
                           break;  
                       }
                   }
                
                   if (!xaresinthread) {
                      break;   
                   }
               }

               // checkin the XAResource, this will "unlock" the XAResource
               myrmreg.rmCheckinXARes();
               break;
           }
        }
       
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("endthread for processResourceManager");
        }
    }
   
    // private void settxxidrecoveraction (XAResource actxares, Xid [] actionxid) {
    private void settxxidrecoveraction(XAResource actxares, List actionxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("settxxidrecoveraction");
        }
    
        TxxidRecovered mytxxidRecovered = null;
        TxRecovered mytxRecovered = null;

        // for each XID (actionxid) returned for a specific XAResource
        // determine if the XID needs to be marked for commit (1)
    
        // for (int i=0; i<actionxid.length; i++) {
        Iterator iter = actionxid.iterator();
        while (iter.hasNext()) {
            boolean xidfound = false;
        //     Xid myjavaxid = actionxid[i];
            Xid myjavaxid = (Xid) iter.next();
            org.objectweb.jotm.Xid myxid = new XidImpl(myjavaxid);
            byte [] mybrqu = myxid.getBranchQualifier();
            
            for (int j=0; j<txrecovered.size(); j++){
                mytxRecovered = (TxRecovered) txrecovered.elementAt(j);
                
                if (mytxRecovered != null) {
                    // We will set every XID found to rollback, if we find
                    // the XID in our tables we will reset the XID to commit
                    // or not one of ours
                
                    for (int k=0; k<mytxRecovered.getxidcount(); k++){
                        mytxxidRecovered = mytxRecovered.getRecoverTxXidInfo(k);
                    
                        if (mytxxidRecovered != null) {
                            byte [] mytxxid = mytxxidRecovered.getRecoverxid();
                            org.objectweb.jotm.Xid mymytxxid = new XidImpl(mytxxid);
                        
                            if (mymytxxid.isThisOneOfOurs(mybrqu)) {
                                
                                if (TraceTm.recovery.isDebugEnabled()) {
                                    TraceTm.recovery.debug("mymytxxid= " + mymytxxid.toString(true));
                                    TraceTm.recovery.debug("myxid  = " + myxid.toString(true));
                                }
                    
                                if ((new String(mytxxid)).equals(myxid.toString(true))){
                                    // reset the RecoveryTxXidInfo entry with the registered
                                    // XAResource and the commit action
                                    xidfound = true;
                                    int myaction = 1;    // Commit
                                    mytxxidRecovered.setRecoveraction(myaction);
                                    mytxxidRecovered.setCommitxares(actxares);
                                    mytxxidRecovered.setCommitxid(myjavaxid);
                                    mytxRecovered.setRecoverTxXidInfo(mytxxidRecovered,k);
                                    break;
                                }
                            } else {
                                if (TraceTm.recovery.isDebugEnabled()) {
                                    TraceTm.recovery.debug("Xid is not one of ours");
                                }
                            }
                        }
                    }
                }
            
                if (xidfound) {    // xid found in a TxxidRecovered of the TxRecovered vector
                    break;
                }
            }
        
            if (!xidfound) {    // xid NOT found in a TxxidRecovered of the TxRecovered vector
                abortimmediate (actxares, myjavaxid);
            }
        }
    }

    private void doActionXidRecover () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("doActionXidRecover");
        }
    
        TxxidRecovered mytxxidRecovered = null;
        TxRecovered mytxRecovered = null;
    
        for (int i=0; i<txrecovered.size(); i++){
            mytxRecovered = (TxRecovered) txrecovered.elementAt(i);
            XAResource commitxares = null;
            Xid commitxid = null;
      
            for (int j=0; j<mytxRecovered.getxidcount(); j++){
                mytxxidRecovered = mytxRecovered.getRecoverTxXidInfo(j);
        
                if (mytxxidRecovered != null) {

                    // If action = 0; XAResource is known but the XID associated
                    // to is was not returned in the xares.recover call. This implies
                    // that the XID was already committed, just ignore.
                    // If action = 1; XAResource is known and the xares.recover call
                    // returned the specifed XID. This implies that we should attempt
                    // to commit the XID (if possible).
            
                    if (mytxxidRecovered.getRecoveraction() == 1) {  // Commit
                        byte [] myrcxid = mytxxidRecovered.getRecoverxid();
                        commitxares = mytxxidRecovered.getCommitxares();
                        commitxid = mytxxidRecovered.getCommitxid();

                        commitXAResourceXid committhread = new commitXAResourceXid (commitxares, commitxid);
                        Thread commitxa = new Thread (xaResourceThreadGroup, committhread, "commitxid");
                        commitxa.start();
                    }
                }
            }
        }
    }

    private void abortimmediate (XAResource abortxares, Xid abortxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("abortimmediate");
        }
    
        abortXAResourceXid abortthread = new abortXAResourceXid (abortxares, abortxid);
        Thread abortxa = new Thread (xaResourceThreadGroup, abortthread, "abortxid");
        abortxa.start();
    }
}

/**
 * Thread class used to commit an Xid of an XAResource,
 * XAResource.commit(Xid).
 */

class commitXAResourceXid extends Thread {
    XAResource commitxares = null;
    Xid commitxid = null;
    
   commitXAResourceXid (XAResource pxares, Xid pxid) {
       if (TraceTm.recovery.isDebugEnabled()) {
           TraceTm.recovery.debug("pxares= " + pxares);
           TraceTm.recovery.debug("pxid= " + pxid);
       }
       
       commitxares = pxares;
       commitxid = pxid;
   }
    
   public void run() {
       if (TraceTm.recovery.isDebugEnabled()) {
           TraceTm.recovery.debug("thread for commitXAResourceXid");
           TraceTm.recovery.debug("  Committing xid= " + commitxid);
           TraceTm.recovery.debug("  with XAResource= " + commitxares);
       }

       try {
           commitxares.commit(commitxid, false);
       } catch (XAException e) {
           TraceTm.recovery.error("Unable to commit Xid during Recovery " + e.getMessage());
       }
       
       if (TraceTm.recovery.isDebugEnabled()) {
           TraceTm.recovery.debug("endthread for commitXAResourceXid");
       }
   }
}

/**
 * Thread class used to abort an Xid of an XAResource,
 * XAResource.rollback(Xid).
 */

class abortXAResourceXid extends Thread {
    XAResource abortxares = null;
    Xid abortxid = null;
    
    abortXAResourceXid (XAResource pxares, Xid pxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("pxares= " + pxares);
            TraceTm.recovery.debug("pxid= " + pxid);
        }
        
        abortxares = pxares;
        abortxid = pxid;
    }
    
    public void run() {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("thread for abortXAResourceXid");
            TraceTm.recovery.debug("  Rolling Back xid= " + abortxid);
            TraceTm.recovery.debug("  with XAResource= " + abortxares);
        }
        
        try {
            abortxares.rollback(abortxid);
        } catch (XAException e) {
            TraceTm.recovery.error("Unable to rollback Xid during Recovery " + e.getMessage());
        }
        
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("endthread for abortXAResourceXid");
        }
    }
}
/**
 * Vector used to hold all Resource Managers that were
 * registered when the system crashed, may have XIDs that
 * need to be recovered.
 */

class RecoverRmInfo {

    private String recoverRm = null;
    private byte[] recoverXares = null;
    private String recoverxaresName = null;
    private int recoverIndex = 0;

    public void addRecoverRmXaRes (String rrmName, byte[] rrmXares, String rrmxaresName, int rrmIndex) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("RecoverRm Resource Manager= " + rrmName);
            TraceTm.recovery.debug("RecoverRm XAResource = " + new String(rrmXares));
            TraceTm.recovery.debug("RecoverRm Index = " + rrmIndex);
        }

        recoverRm = rrmName;
        recoverXares = rrmXares;
        recoverxaresName = rrmxaresName;
        recoverIndex = rrmIndex;
    }

    public String getRecoverRm () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverRm= " + recoverRm);
        }
        return recoverRm;
    }

    public byte[] getRecoverXaRes () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverXares= " + new String(recoverXares));
        }
        return recoverXares;
    }
    
    public String getRecoverXaResName () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverXares= " + recoverxaresName);
        }
        return recoverxaresName;
    }
    
    public int getRecoverIndex () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverIncex= " + recoverIndex);
        }
        return recoverIndex;
    }
}

/**
 * Vector and Array used to hold all Transactions that were
 * in a commit pending status when the system crashed, these
 * transactions (XIDs) may need to be recovered.
 */

class TxxidRecovered {

    private int txxidIndex = 0;
    private byte[] txxidXares = null;
    private String txxidXaresname = null;
    private byte[] txxidXid = null;
    private int txxidStatus = 0;
    private int txxidAction = 0;    // 0-Do nothing, implied already committed
                                    // 1-Commit
    private XAResource commitxares = null;
    private Xid commitxid = null;

    public void addXidInfo (int pindex, byte[] pxares, String pxaresname, byte[] pxid, int pstatus) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Txxid index= " + pindex);
            TraceTm.recovery.debug("Txxid xares= " + new String(pxares));
            TraceTm.recovery.debug("Txxid xid= " + new String(pxid));
            TraceTm.recovery.debug("Txxid status= " + pstatus);
        }

        txxidIndex = pindex;
        txxidXares = pxares;
        txxidXaresname = pxaresname;
        txxidXid = pxid;
        txxidStatus = pstatus;
    }

    public int getRecoverindex () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidIndex= " + txxidIndex);
        }
        return txxidIndex;
    }

    public byte[] getRecoverxares () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidXares= " + new String(txxidXares));
        }
        return txxidXares;
    }

    public String getRecoverxaresname () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidXaresname= " + txxidXaresname);
        }
        return txxidXaresname;
    }

    public byte[] getRecoverxid () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidXid= " + new String(txxidXid));
        }
        return txxidXid;
    }

    public int getRecoverstatus () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidStatus= " + txxidStatus);
        }
        return txxidStatus;
    }
    
    public void setRecoverstatus (int pstatus) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("pstatus= " + pstatus);
        }
        txxidStatus = pstatus;
    }
    
    public int getRecoveraction () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("txxidAction= " + txxidAction);
        }
        return txxidAction;
    }
    
    public void setRecoveraction (int paction) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("paction= " + paction);
        }
        txxidAction = paction;
    }
    
    public XAResource getCommitxares () {
        return commitxares;
    }
    
    public void setCommitxares (XAResource pcommitxares) {
        commitxares = pcommitxares;
    }
    
    public Xid getCommitxid () {
        return commitxid;
    }
    
    public void setCommitxid (Xid pcommitxid) {
        commitxid = pcommitxid;
    }
}

class TxRecovered {
    private long recoverydatetime = 0L;
    private byte[] txxid = null;
    private String txdatetime = null;
    private int xidcount = 0;
    private XACommittingTx xacommittingtx = null;

    private TxxidRecovered [] xidinfo;

    public void addtxrecovered (long prdt, byte[] ptxxid, String ptdt, int pxcnt, XACommittingTx pxacmtx) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recover tx prdt= " + prdt);
            TraceTm.recovery.debug("Recover tx ptxxid= " + new String(ptxxid));
            TraceTm.recovery.debug("Recover tx ptdt= " + ptdt);
            TraceTm.recovery.debug("Recover tx pxcnt= " + pxcnt);
            TraceTm.recovery.debug("Recover tx pxacmtx= " + pxacmtx);
        }

        recoverydatetime = prdt;
        txxid = ptxxid;
        txdatetime = ptdt;
        xidcount = pxcnt;
        xacommittingtx = pxacmtx;

        xidinfo = new TxxidRecovered [xidcount];
    }

    public void setRecoverTxXidInfo (TxxidRecovered ptxxidr, int rindx) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("addRecoverTxXidInfo");
        }
        xidinfo[rindx] = ptxxidr;
    }
    
    public long getrecoverdatetime () {
        return recoverydatetime;
    }
    
    public byte [] gettxxid () {
        return txxid;
    }
    
    public String gettxdatetime () {
        return txdatetime;
    }
    
    public int getxidcount () {
        return xidcount;
    }
    
    public TxxidRecovered getRecoverTxXidInfo (int rindx) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("getRecoverTxXidInfo");
        }
        return xidinfo[rindx];
    }
    
    public XACommittingTx getXACommittingTx () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("getXACommittingTx");
        }
        return xacommittingtx;
    }
}

public class JotmRecovery {
    
    private static JotmRecovery unique = null;

    private static Vector vTxRecovered = new Vector ();
    private static Vector vRecoverRmInfo = new Vector ();
    private Vector userRecoveryRecords = new Vector ();
    
    /**
     * Constructor.
     */

    public JotmRecovery() {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("JotmRecovery constructor");
        }

        unique = this;
    }
    
    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized.
     *
     * @return The <code>JotmRecovery</code> object created 
     */

    public static JotmRecovery getJotmRecovery() {

        return unique;
    }
    
    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized.
     *
     * @return The <code>TxRecovered</code> vector created 
     */
    public static Vector getTxRecovered() {
        return vTxRecovered;
    }
    
    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized.
     *
     * @return The <code>RecoverRmInfo</code> vector created 
     */
    public static Vector getRecoverRmInfo() {
        return vRecoverRmInfo;
    }

    public Vector getmyRecoverRmInfo() {
        return vRecoverRmInfo;
    }

    /**
     * Returns the index of the Resource Manager's XAResource.
     *
     * @return Index of the Resource Manager's XAResource. 
     */
    public int getRmIndex(byte [] pxares) {
        int numRm = vRecoverRmInfo.size();
        RecoverRmInfo myrecoverRmInfo;
        
        for (int i=0; i<numRm; i++){
            myrecoverRmInfo = (RecoverRmInfo) vRecoverRmInfo.elementAt(i);
            byte [] inrmxares = myrecoverRmInfo.getRecoverXaRes();
            
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("XAResource param " + Arrays.toString(pxares));
                TraceTm.recovery.debug("XAResource in rm " + Arrays.toString(inrmxares));
            }

            if (inrmxares == pxares) {
                return myrecoverRmInfo.getRecoverIndex();
            }
        }    
        return 99;
    }

    public Vector getUserRecoveryVector() { // currently used by test suite
        return userRecoveryRecords;
    }
    
    /**
     * Processes an XACOMMIT entry (putCommit) that does not have an associated
     * XADONE entry (putDone).
     *
     * @param lr LogRecord that was passed to onRecord() method.
     */
    public void rebuildTransaction (XALogRecord lr) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("rebuildTransaction");
        }
        
        RecoverRmInfo myrecoverRmInfo;
        TxxidRecovered myrecoverTxInfo;
        TxRecovered mytxRecovered;
        
        // Temporary Recovery Record
        byte [] tempRec;
        byte [] rt = new byte [3];  // Applies to all Records

        // Resource Manager Record Type 1 fields
        int rmcount;

        // Resource Manager Record Type 2
        byte [] resmgr2;

        // Resource Manager Record Type 2 fields
        int rmlength;
        byte [] rmname;
        int rmindx;

        // Recovery Record Type 1 fields
        long rcdatetime;
        int txxidlength;
        byte [] txXid;
        int txdatelength;
        byte [] txdatetime;
        int xarescount = 0;

        // Check if there already exist a recovery record in our array with the
        // same txXid but a older datetime. If so remove it and store this one,
        // otherwise throw this one away.

        // Recovery Record Type 2
        byte [] recov2;

        // Recovery Record Type 2 fields
        byte [] rt2 = new byte [3];
        int xaresindex;
        int xareslength;
        int xaresnamelength;
        byte [] xares;
        int xidlength;
        byte [] xaresname;
        byte [] recoveryxid;
        int xidstatus;

        String trt;
        
        XACommittingTx myxacommittx = lr.getTx();
        tempRec = lr.getFields() [0];

        ByteBuffer rr = ByteBuffer.wrap(tempRec);

        rr.get(rt, 0, 3);
        trt = new String(rt);

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recovery Record type= " + trt);
        }

        if (trt.equals("RM1")) {

            Long rmdatetime = rr.getLong();
            rmcount = rr.getInt();

            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Resource Manager count= " + rmcount);
            }

            for (int i=1; i <= rmcount; i++) {
                resmgr2 = lr.getFields()[i];

                ByteBuffer resm2 = ByteBuffer.wrap(resmgr2);
                resm2.get(rt2, 0, 3);
                trt = new String (rt2);

                if (trt.equals("RM2")) {
                    rmlength = resm2.getInt();
                    rmname = new byte[rmlength];
                    resm2.get(rmname, 0, rmlength);
                    xareslength = resm2.getInt();
                    xares = new byte[xareslength];
                    resm2.get(xares, 0, xareslength);
                    xaresnamelength = resm2.getInt();
                    xaresname = new byte[xaresnamelength];
                    resm2.get(xaresname, 0, xaresnamelength);
                    rmindx = resm2.getInt();

                    String myrmname = new String(rmname);
                    String myxaresname = new String(xaresname);
                    myrecoverRmInfo = new RecoverRmInfo();

                    myrecoverRmInfo.addRecoverRmXaRes (myrmname, xares, myxaresname, rmindx);
                    vRecoverRmInfo.addElement(myrecoverRmInfo);
                }
            }
        }  else if (trt.equals("RR1")) {
                rcdatetime = rr.getLong();
                txxidlength = rr.getInt();
                txXid = new byte[txxidlength];
                rr.get(txXid, 0, txxidlength);
                txdatelength = rr.getInt();
                txdatetime = new byte[txdatelength];
                rr.get(txdatetime, 0, txdatelength);
                xarescount = rr.getInt();

                String mytxdatetime = new String(txdatetime);

                mytxRecovered = new TxRecovered();

                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("XAResource count= " + xarescount);
                }

                mytxRecovered.addtxrecovered (rcdatetime, txXid, mytxdatetime, xarescount, myxacommittx);
                
                for (int i = 1; i <= xarescount; i++) {
                    myrecoverTxInfo = new TxxidRecovered();
                    recov2 = lr.getFields()[i];

                    ByteBuffer rr2 = ByteBuffer.wrap(recov2);
                    rr2.get(rt2, 0, 3);
                    trt = new String (rt2);

                    if (trt.equals("RR2")) {
                        xaresindex = rr2.getInt();
                        xareslength = rr2.getInt();
                        xares = new byte[xareslength];
                        rr2.get(xares, 0, xareslength);
                        xaresnamelength = rr2.getInt();
                        xaresname = new byte[xaresnamelength];
                        rr2.get(xaresname, 0, xaresnamelength);
                        xidlength = rr2.getInt();
                        recoveryxid = new byte[xidlength];
                        rr2.get(recoveryxid, 0, xidlength);
                        xidstatus = rr2.getInt();

                        String myxaresname = new String(xaresname);

                        myrecoverTxInfo.addXidInfo (xaresindex, xares, myxaresname, recoveryxid, xidstatus);
                        mytxRecovered.setRecoverTxXidInfo (myrecoverTxInfo, i-1);
                    }
                }
                vTxRecovered.addElement(mytxRecovered);
        } else if (trt.equals("RU1")) {
                XidImpl.setUuids( rr.getLong(), rr.getLong());
        } else {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Unknown record type during replay = " + trt);
            }
            // add this byteBuffer of this record to a vector for user recovery
            rr.rewind();
            userRecoveryRecords.add( rr );
            userRecoveryRecords.add( myxacommittx );
        }
    }

    public void recoverTransactions (Vector rmreg) throws XAException {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverTransactions");
        }

        int rmregsize = rmreg.size();  // number of Resource Managers registered

        // read the tmRecovered object and determine if any of the xaResources
        // require recovery (xares.recover call)

        int rmsize = vRecoverRmInfo.size();    // number of Resource Managers when system crashed

        if (rmsize == 0) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Nothing to recover");
            }
            return;
        }

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("number Resource Manager recover= " + rmsize);
        }

        // we can only recover Resource Managers that are in our log
        // (RecoverRmInfo) and have been registered (RmRegistered)
        
        ThreadGroup resourceManagerThreadGroup = new ThreadGroup("ResourceManagerTG");
        
        for (int i = 0; i < rmregsize; i++) {
            RmRegistration myrmreg = (RmRegistration) rmreg.elementAt(i);
            String myregrm = myrmreg.rmGetName();
            processResourceManager rmthread = new processResourceManager (resourceManagerThreadGroup, vTxRecovered, vRecoverRmInfo, rmreg, rmsize, i);
            Thread processrm = new Thread (resourceManagerThreadGroup, rmthread, myregrm);
            processrm.start();
        }
        
        for (int w=0; w<10; w++) {  // only do it 10 times)
            boolean rmsinthread = false;
            int numThreads = resourceManagerThreadGroup.activeCount();
            
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("resourceManager active count= "+ numThreads);
            }
            
            if (numThreads == 0) {
                break;
            }
                
            Thread[] listOfThreads = new Thread[numThreads];
            resourceManagerThreadGroup.enumerate(listOfThreads);

            for (int i = 0; i < numThreads; i++) {
                 if (listOfThreads[i] != null) {
                     if (TraceTm.recovery.isDebugEnabled()) {
                         TraceTm.recovery.debug("resourceManager in thread= "+ listOfThreads[i].getName());
                     }
                     
                     try {
                         Thread.sleep(1000);  // 1 second
                     } catch (InterruptedException e) {
                         ;
                     }
                     rmsinthread = true;
                     break;  
                }
            }
            
            if (!rmsinthread) {
                break;   
            }
        }
        
        // We can now walk through TxRecovered and delete (howl done) any TxRecovered
        // entries that have had all their Xids (TxxidRecovered) committed or rolled
        // back successfully.
        // If any Xids have not been committed or rolledback, create a new TxRecovered
        // entry (howl commit) with the Xids (TxxidRecovered) that are still pending.

        doCleanupXidRecover ();
        
        for (int i=0; i<rmregsize; i++) {
            RmRegistration myrmreg = (RmRegistration) rmreg.elementAt(i);
            myrmreg.rmSetRmRecovered(true);
        }
    }
    
    private void doCleanupXidRecover () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("doCleanupXidRecover");
        }
        
        TxxidRecovered myTxxidRecovered = null;
        TxRecovered mytxRecovered = null;
        
        // We must go upward (last element to first) since we may remove
        // elements of the vTxRecovered vector.
        
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("vTxRecovered.size= " +vTxRecovered.size());
        }
        
        for (int i=vTxRecovered.size()-1; i>=0; i--) {
            boolean possibleheuristic = false;
            XACommittingTx myxacommittingtx = null;
            mytxRecovered = (TxRecovered) vTxRecovered.elementAt(i);
            
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("mytxRecovered.xidcount= " +mytxRecovered.getxidcount());
            }
            
            for (int j=0; j<mytxRecovered.getxidcount(); j++){
                myTxxidRecovered = mytxRecovered.getRecoverTxXidInfo(j);
                
                if (myTxxidRecovered != null) {
                    if ((myTxxidRecovered.getRecoveraction() == 0) ||  // Xid already committed, ignore
                        (myTxxidRecovered.getRecoveraction() == 1)) {  // Xid just committed
                        ;
                    } else {
                        if (TraceTm.recovery.isDebugEnabled()) {
                            TraceTm.recovery.debug("possibleheuristic");
                        }
                        
                        possibleheuristic = true;
                    }
                }
            }
            
            // We cannot remove the vTxRecovered vector element if there
            // exists the possibility of a heuristic transaction.
            
            if (!possibleheuristic) {

                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("write howlDonelog");
                }
                
            	myxacommittingtx = mytxRecovered.getXACommittingTx();
                byte [] rmDone = new byte [11];
                byte [] [] rmDoneRecord = new byte [1] [11];

                rmDone = "RR3JOTMDONE".getBytes();

                try {
                    rmDoneRecord [0] = rmDone;
                    TransactionRecoveryImpl.getTransactionRecovery().howlDoneLog (rmDoneRecord, myxacommittingtx);
                } catch (Exception f) {
                    String howlerror =
                        "Cannot howlDoneLog:"
                        + f
                        + "--"
                        + f.getMessage();
                    TraceTm.jotm.error("Got LogException from howlDoneLog: "+ howlerror);
                }

                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("remove txRecovered entry");
                }
                
                vTxRecovered.remove(i);
            }
        }
    }
}